////////////////////////////////////////
/**************************************/
/*                                    */
/* Michael Goldbuch; Jrgen Dffinger */
/*                                    */
/* Mikroprozessortechnik              */
/* Linienverfolgung		              */
/* Version 1.0.0                      */
/* 2011-06-09                         */
/*                                    */
/**************************************/
////////////////////////////////////////



//	definitions for ATMega8
.include "m8def.inc"



/**************** Definitionen   *****************/

.DSEG

PDR1:	.Byte	8
PDR2:	.Byte	8

.CSEG

.def  adcc   	=  	R10	; ADC Zhler
.def  zel    	=  	R11	; Zwischenergebnis rechter Sensor
.def  zer    	=  	R12	; Zwischenergebnis linker Sensor
.def  ml     	=  	R13	; Messwert linker Sensor
.def  mr     	=  	R14	; Messwert rechter Sensor
.def  diff   	=  	R15 	; Differenz von Messwert linker Sensor minus Messwert rechter Sensor
.def  sr	 	=  	R8
.def  sl     	=  	R9
.equ  wartezeit = 	250
.equ  drift	    = 	252
.equ  messwert  = 	0


/***************   Interupts   ***************/

rjmp RESET ; Reset Handler
reti; rjmp EXT_INT0 ; IRQ0 Handler
reti; rjmp EXT_INT1 ; IRQ1 Handler
reti; rjmp TIM2_COMP ; Timer2 Compare Handler
reti; rjmp TIM2_OVF ; Timer2 Overflow Handler
reti; rjmp TIM1_CAPT ; Timer1 Capture Handler
reti; rjmp TIM1_COMPA ; Timer1 CompareA Handler
reti; rjmp TIM1_COMPB ; Timer1 CompareB Handler
reti; rjmp TIM1_OVF ; Timer1 Overflow Handler
rjmp TIM0_OVF ; Timer0 Overflow Handler
reti; rjmp SPI_STC ; SPI Transfer Complete Handler
reti; rjmp USART_RXC ; USART RX Complete Handler
reti; rjmp USART_UDRE ; UDR Empty Handler
reti; rjmp USART_TXC ; USART TX Complete Handler
rjmp ADCI ; ADC Conversion Complete Handler
reti; rjmp EE_RDY ; EEPROM Ready Handler
reti; rjmp ANA_COMP ; Analog Comparator Handler
reti; rjmp TWSI ; Two-wire Serial Interface Handler
reti; rjmp SPM_RDY ; Store Program Memory Ready Handler

.include "arithmetic.asm"
.include "USART.asm"


/***************   RESET   ***************/

RESET:

	// 	Stack initialisieren
	ldi R16,high(RAMEND)
	out sph,R16
	ldi R16,low(RAMEND)	
	out spl,R16	


	//	USART initialisieren
	clr R16													; Set baud rate (2400 BAUD)

	ldi R17,0xD0

	out UBRRH, R16

	out UBRRL, R17

	ldi R16, (1<<URSEL)|(1<<USBS)|(3<<UCSZ0)				; Set frame format: 8data, 2stop bit

	out UCSRC,R16
	sbi UCSRB,TXEN


	
	//	ADC initialisieren
	ldi  R16, (1 << REFS0)|(1 << ADLAR)|(1 << MUX1)
	out  ADMUX,R16
	ldi  R16, (1 << ADEN)|(1 << ADPS2);|(1 << ADPS1) | (1 << ADPS0)  	; Prescaler 128
	out  ADCSRA, R16
	cbi  DDRC,2
	cbi  DDRC,3
	clr  adcc
	inc  adcc
	clr  ml
	clr  mr
	clr  zel
	clr  zer
	clr  diff

	
	//	Motoren initialisieren
	in  R16,DDRB											; Ports initialisieren
	ori R16,0b00110110
	out DDRB,R16
	in  R16,DDRD
	ori R16,0b00110000
	out DDRD,R16
	sbi PORTB,1
	sbi PORTB,2
	ldi R16, (1 << WGM10) | (1 << COM1A1) | (1 << COM1B1)	; for PWM (8-Bit PWM) on OC1A & OC1B
	out TCCR1A, R16
	ldi R16, (1 << CS11)
	out TCCR1B, R16


	//	LEDs initialisieren
	sbi DDRD,7
	cbi PORTD,7
	sbi DDRC,0
	sbi DDRC,1
	cbi PORTC,0
	cbi PORTC,1
	sbi DDRB,0
	cbi PORTB,0
	sbi DDRD,2
	cbi DDRD,2
	sbi DDRD,6
	sbi PORTD,6

	// Timer initialisieren
	ldi R16,(1 << CS02)|(1 << CS00)
	out TCCR0,R16
	
	sei	; global Interrupt enable

rjmp main



/*************** main   ***************/

main:

	//	Tabelle (PD-Regler) anlegen
	ldi ZH,high(PDR1)
	ldi ZL,low(PDR1)
	ldi R18,160
	st  Z+,R18
	ldi R18,105
	st  Z+,R18
	ldi R18,72
	st  Z+,R18
	ldi R18,51
	st  Z+,R18
	ldi R18,39
	st  Z+,R18
	ldi R18,31
	st  Z+,R18
	ldi R18,27
	st  Z+,R18
	ldi R18,20
	st  Z+,R18

	ldi ZH,high(PDR2)
	ldi ZL,low(PDR2)
	ldi R18,160
	st  Z+,R18
	ldi R18,125
	st  Z+,R18
	ldi R18,106
	st  Z+,R18
	ldi R18,96
	st  Z+,R18
	ldi R18,91
	st  Z+,R18
	ldi R18,88
	st  Z+,R18
	ldi R18,87
	st  Z+,R18
	ldi R18,85
	st  Z+,R18


	//	Startwerte setzen
	ldi R16,0x01
	mov mr,R16
	mov ml,R16

	//	Timer starten
	ldi R16,wartezeit
	out TCNT0,R16
	ldi R16,(1 << TOIE0)
	out TIMSK,R16

	ldi R18,messwert
	tst R18
	brne main_1

	// Motor vorwrtsfahrt initialisieren
	sbi PORTD,5
	cbi PORTD,4
	sbi PORTB,5
	cbi PORTB,4
	ldi R18,80
	mov sr,R18
	mov sl,R18
	rjmp Motor_Speed	

main_1:

		
	tst mr
	brne main_2
	tst ml
	brne main_2
	rcall Motor_Stop
	main_1_1:
	rjmp main_1_1

main_2:
	tst  diff
	brmi rechts_fahren
	rjmp links_fahren

	rechts_fahren:

	ser  R18
	sub  R18,diff
	inc  R18
	cpi  R18,7
	brcs rechts_fahren_1
	ldi  R18,7

	rechts_fahren_1:

	ldi  ZH,high(PDR1)
	ldi  ZL,low(PDR1)
	add  ZL,R18
	ld   R19,Z
	mov  sr,R19
	ldi  ZH,high(PDR2)
	ldi  ZL,low(PDR2)
	add  ZL,R18
	ld   R19,Z
	mov  sl,R19
	rjmp Motor_Speed

	links_fahren:

	mov  R18,diff
	cpi  R18,7
	brcs links_fahren_1
	ldi  R18,7

	links_fahren_1:

	ldi  ZH,high(PDR1)
	ldi  ZL,low(PDR1)
	add  ZL,R18
	ld   R19,Z
	mov  sl,R19
	ldi  ZH,high(PDR2)
	ldi  ZL,low(PDR2)
	add  ZL,R18
	ld   R19,Z
	mov  sr,R19
	rjmp Motor_Speed

	Motor_Speed:
	
	ldi R18,messwert
	tst R18
	brne main_tx

	clr R16
	out OCR1AH, R16
	out OCR1AL, sl
	out OCR1BH, R16
	out OCR1BL, sr

rjmp main_1


main_tx:

	TXI 'L'
	TXI ':'
	TX_BIN2ASCII R25,ml
	TXI 'R'
	TXI ':'
	TX_BIN2ASCII R25,mr
	TXI 'D'
	TXI ':'
	TX_BIN2ASCII R25,diff
	TXI 'S'
	TXI 'L'
	TXI ':'
	TX_BIN2ASCII R25,sl
	TXI 'S'
	TXI 'R'
	TXI ':'
	TX_BIN2ASCII R25,sr
	TXI 10
	TXI 13

rjmp main_1






/***************   ADCI   ****************/

adci:

	push R16
	in   R16,SREG
	push R16
	push R17
	push R28

	adci_0:	
	
	in   R28,ADCH
	mov  R17,adcc	; Zhler kopieren
	lsr  R17		; auslesen des ersten Bits
	brcc adci_1		

	//	Wenn erstes Bit des Zhlers gesetzt ist, dann Messwert linker Sensor mit LED
	//	zwischenspeichern und ADC auf Messung linker Sensor ohne LED vorbereiten

	lsl  adcc		; ADC Zhler setzen fr zweite Messung
	mov  zel,R28	; Messwert in Zwischenspeicher (zel - Zwischenergebnis links) speichern
	cbi  PORTD,6	; Front LED abschalten
	rjmp adci_4

	adci_1:

	lsr  R17		; auslesen zweites Bit des Zhlers
	brcc adci_2		

	//	Wenn zweites Bit gesetzt ist, dann Messwert linker Sensor ohne LED vom
	//	Zwischengespeicherten Messwert mit LED abziehen und zwischenspeichern
	//	sowie ADC fr Messung rechten Sensor mit LED vorbereiten
	
	lsl  adcc		; ADC Zhler setzen fr dritte Messung
	sub  zel,R28	; Messwert von Zwischenergebnis abziehen
	ldi  R16, (1 << REFS0)|(1 << ADLAR)|(1 << MUX1)|(1 << MUX0)	; ADC auf rechten Sensor einstellen
	out  ADMUX,R16
	sbi  PORTD,6	; Front LED einschalten
	rjmp adci_4

	adci_2:

	lsr  R17		; auslesen drittes Bit des Zhlers
	brcc adci_3

	//	Wenn drittes Bit des Zhlers gesetzt ist, dann Messwert rechter Sensor mit LED zwischenspeichern
	//	und ADC auf Messung rechter Sensor ohne LED vorbereiten

	lsl  adcc		; ADC Zhler setzen fr vierte Messung
	mov  zer,R28	; Messwert in Zwischenspeicher (zer - Zwischenergebnis rechts) speichern
	cbi  PORTD,6	; Front LED abschalten
	rjmp adci_4

	adci_3:

	//	Wenn auch drittes Bit nicht gesetzt war, dann wurde vierte Messung durchgefhrt.
	//	Dazu Messwert rechter Sensor ohne LED von Zwischenergebnis Messung rechter Sensor mit LED abziehen 
	//	und zwischenspeichern, anschlieend Auswertung aller Zwischenergebnisse und in den
	//	entsprechenden Registern die Ergebnisse speichern

	sub  zer,R28	; Messwert von Zwischenergebnis abziehen
	ldi  R16, (1 << REFS0)|(1 << ADLAR)|(1 << MUX1)	; ADC auf linken Sensor einstellen
	out  ADMUX,R16
	sbi  PORTD,6	; Front LED einschalten
	mov  ml,zel		; Zwischenergebnis links als Messwert links speichern
	mov  mr,zer		; Zwischenergebis rechts als Messwert rechts speichern
	mov  diff,ml	; Differenz bilden (Messwert links - Messwert rechts)
	sub  diff,mr
	ldi  R17,drift		
	add  diff,R17
	ldi  R16,1
	mov  adcc,R16
	clr  R16
	
	adci_4:  
	
	cbi ADCSRA,ADIE
	ldi R16,wartezeit
	out TCNT0,R16
	ldi R16,(1 << TOIE0)
	out TIMSK,R16

	pop  R28
	pop  R17
	pop  R16
	out  SREG,R16
	pop  R16

reti



/*************** Timer   ***************/

TIM0_OVF:

	push R16
	in   R16,SREG
	push R16
	
	// Timer abschalten
	in   R16,TIMSK
	andi R16,0b11111110
	out  TIMSK,R16

	// ADC einschalten
	sbi ADCSRA,ADIE
	sbi ADCSRA,ADSC

	pop R16
	out SREG,R16
	pop R16

reti



/****************   Motor_Stop   ***************/

Motor_Stop:

	push R16
	
	clr R16
	out OCR1AH, R16
	out OCR1AL, R16
	out OCR1BH, R16
	out OCR1BL, R16

	pop R16

ret
